<?php

namespace NoccyLabs\MessageCatalog;

use PDO;
use Traversable;

class MessageCatalog
{
    private PDO $pdo;

    public function __construct(private string $filename)
    {
        $this->pdo = new PDO("sqlite:{$filename}");
        $this->pdo->exec("CREATE TABLE IF NOT EXISTS messages (".
            "id integer primary key autoincrement,".
            "collection varchar not null,".
            "name varchar not null".
        ")");
        $this->pdo->exec("CREATE TABLE IF NOT EXISTS contents (".
            "id integer primary key autoincrement,".
            "message_id integer,".
            "type varchar not null,".
            "content blob not null,".
            "foreign key(message_id) references messages(id),".
            "unique(message_id, type)".
        ")");
    }

    private function pdoQuery(string $query, array $params=[]): Traversable
    {
        $stmt = $this->pdo->prepare($query);
        if (!$stmt->execute($params))
            throw new \RuntimeException();
        while ($row = $stmt->fetch(PDO::FETCH_ASSOC)) {
            yield $row;
        }
    }

    private function pdoQueryOne(string $query, array $params=[]): ?array
    {
        $stmt = $this->pdo->prepare($query);
        if (!$stmt->execute($params))
            throw new \RuntimeException();
        return $stmt->fetch(PDO::FETCH_ASSOC);
    }

    private function pdoQuerySingleValue(string $query, array $params=[]): mixed
    {
        $stmt = $this->pdo->prepare($query);
        if (!$stmt->execute($params))
            throw new \RuntimeException();
        $row = $stmt->fetch(PDO::FETCH_ASSOC);
        if ($row === false) return null;
        return array_shift($row);
    }

    private function pdoExec(string $query, array $params=[]): void
    {
        $stmt = $this->pdo->prepare($query);
        if (!$stmt->execute($params))
            throw new \RuntimeException();
    }

    public function getMessage(string $collection, string $name): Message
    {
        $match = (array)$this->pdoQueryOne(
            "SELECT * FROM messages WHERE collection=:collection AND name=:name", [
                'collection' => $collection,
                'name' => $name,
            ]);
        
        if(!$match)
            throw new \RuntimeException();

        $contents = [];
        foreach ($this->pdoQuery("SELECT * FROM contents WHERE message_id=:id", [ 'id'=>$match['id']]) as $row) {
            $contents[$row['type']] = $row['content'];
        }

        // $defaults = (array)$this->pdoQuery("SELECT * FROM defaults WHERE (valid_from IS NULL OR datetime(:now) >= datetime(valid_from)) AND (valid_to IS NULL OR datetime(:now) < datetime(valid_to))");

        $message = new Message(...$contents);

        return $message;
    }

    public function findMessages(?string $collection = null): array
    {
        if ($collection !== null) {
            $hits = $this->pdoQuery("SELECT * FROM messages WHERE collection=:collection ORDER BY name ASC", [ 'collection' => $collection]);
        } else {
            $hits = $this->pdoQuery("SELECT * FROM messages ORDER BY collection ASC, name ASC", []);
        }

        $result = [];
        foreach ($hits as $row) {
            $result[] = [
                'collection'=>$row['collection'],
                'name'=>$row['name']
            ];
        }

        return $result;
    }

    public function setMessage(string $collection, string $name, Message $message): void
    {
        $messageId = $this->pdoQuerySingleValue(
            "SELECT id FROM messages WHERE collection=:collection AND name=:name", [
                'collection' => $collection,
                'name' => $name,
            ]);
        if (!$messageId) {
            // insert and set $messageId
            $this->pdoExec(
                "INSERT INTO messages (collection,name) VALUES (:collection, :name)", [
                    'collection' => $collection,
                    'name' => $name,
                ]);
            $messageId = $this->pdo->lastInsertId("messages");
        }

        foreach ($message->getAllContent() as $type=>$content) {
            if ($content === null) {
                $this->pdoExec(
                    "DELETE FROM contents WHERE message_id=:message_id AND type=:type", [
                        'message_id' => $messageId,
                        'type' => $type,
                    ]);        
            } else {
                $current = $this->pdoQuerySingleValue(
                    "SELECT id FROM contents WHERE message_id=:message_id AND type=:type", [
                        'message_id' => $messageId,
                        'type' => $type,
                    ]);
                if (!$current) {
                    $this->pdoExec(
                        "INSERT INTO contents (message_id, type, content) VALUES (:message_id, :type, :content)", [
                            'message_id' => $messageId,
                            'type' => $type,
                            'content' => $content,
                        ]);
                } else {
                    $this->pdoExec(
                        "UPDATE contents SET content=:content WHERE message_id=:message_id AND type=:type", [
                            'message_id' => $messageId,
                            'type' => $type,
                            'content' => $content,
                        ]);    
                }
            }
        }
    }

}

